#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <extvab.h>
#include <algorithm>
#include <set>
#include "global.h"
#include "vertex.h"
using namespace std;

int align_bending(char *A, char *B, int M, int N, int** CC, int **S, int *pp);
int align(char *A, char *B, int M, int N, int **S, int *pp);
int traceback(int **S, int M, int N, int *trace, char* A);
int newalign(int *trace, int trace_len, char **aln, int M, char *seq1, int N, char **alnw, int numseq);


int align_bending(char *A, char *B, int M, int N, int**CC, int **S, int *pp)
{
	//implement a bending version of align to reduce the calculation amount.
	//assume that M is approximately the same as N and L is large enough so
	//that makes the bending area likes the following diagram:
/*
		   	B(0,...N)
	<-----L------>
	+--------------------+
A	|					\		|
	|		 			 \	 	|
	|					  \ 	|
0	|\						\ 	|
.	| \					 \	|
.	|	\					  \|
M	| 	 \						|
	|	  \					|
	+--------------------+
*/
	int i, j, k, p;
	int current;
	int direction;
	int temp;
	int score;
	
	int g = 1;
	int h = 0;

	p = 0;

	k=0;
	for (i=1; i < M+1; i++)	
	{
		CC[i][0] = 0;//no cost for going down along the first column[i][0]
		S[i][0]=3; //direction=3, go down
//		k = k+wh[4][A[i]];
//		CC[i][0] = k;
//		if(wh[4][A[i]]) S[i][0] = 2;
//		else S[i][0] = 3;
	}

	for (i=1; i < N+1; i++)	{
		CC[0][i] = gapB*i;
		S[0][i] = 1; 
	}
	CC[0][0] = 0;
	S[0][0] = 0;

	for (i = 0; i < M+1; ++i)
	{
		j = L+i;
		if(j<N+1)	CC[i][j] = oo;
		if(i>L+1)	
		{
			j = i-L;
			if(j>=0 && j<N+1)	CC[i][j] == oo;
		}
	}

	for (i = 1; i < M+1; i++)	{
		for (j = 1; j < N+1; j++)	{
			if(j>=L+i) 
			{
				break;
			}
			if(i>=L+1 && j<=i-L)
			{
				j=i-L;
				continue;
			}
			temp = CC[i-1][j-1];
			current = CC[i-1][j-1] + wh[B[j]][A[i]];

			if (wh[B[j]][A[i]] == 0) direction = 0;
			else direction = -1;
			
			if (CC[i][j-1] + gapB < current)
			{
				current = CC[i][j-1] + gapB;
				direction = 1;
			}

			//if '-' is in set A[i]
			if (wh[4][A[i]] == 0)
			{
				if (CC[i-1][j] < current)	{
					current = CC[i-1][j];
					direction = 3;
				}
			}
			else //'-' is not in set A[i]
			{
				if (CC[i-1][j] + g  < current)
				{
               current = CC[i-1][j] + g;
               direction = 2;
            }
			}
			CC[i][j] = current;
			S[i][j] = direction;
		}
	} 

	*pp = p;
	/*global alignment*/	score = CC[M][N];
   int minrow = 0;
   int minval = oo;
   for (i = M; i>=1 ; i--) {
	    (CC[i][N]<=minval)?(minval=CC[i][N],minrow=i):(0);
   }
   //now we have C[minrow][N] be the minimum score
   score = CC[minrow][N];
	//upadate S[minrow+1][N] to S[M][N];
   for(i=minrow+1; i<M+1; i++) {
      S[i][N]=3;
   }
//	printf("score=%d, minrow=%d, totallength=%d\n",score,minrow, M);
//	printf("     ");
//	for(j=1; j<N+1; j++) {
//		printf("%5d", B[j]);
//	}
//	printf("\n");
//
//	for(i=1; i<M+1;i++) {
//		printf ("%5d", A[i]);
//		for(j=1; j<N+1;j++) {
//			printf ("%5d", CC[i][j]);
//		}
//		printf("\n");
//	}
//	for(i=1; i<M+1; i++) {
//		for(j=1; j<N+1; j++) {
//			if(S[i][j]==-1 ||S[i][j]==0) {
//				printf("\\    ");
//			}
//
//			if(S[i][j]==1) printf("--   ");
//			if(S[i][j]==2 || S[i][j]==3) {
//				printf("|    ");
//			}
//		}
//		printf("\n");
//	}
	return score;
}

int align(char *A, char *B, int M, int N, int **S, int *pp)
{
//allow local alignment (don't penalty gaps in both ends)
/*
	  	B(0,...N)
	+------------+
A	|            |
   |            |
  \|/           |
0	|            |
.	| \          |
.	|   \        |
M	|     \      |
   |       \    |
   |         \  |
	|            |	
	|           \|/	
   |            |
   +------------+ 
*/

	int i, j, k, p;
	int **CC;
	int current;
	int direction;
	int temp;
	int score;
	
	int g = 1;
	int h = 0;
	p = 0;

	CC = (int **) malloc( (M+1)* sizeof (int*));
	for (i = 0; i < M+1 ; i++)	{
		CC[i] = (int *) malloc ( (N+1) * sizeof (int));
	}

	//Boundary cases: 

	k=0;
	for (i=1; i < M+1; i++)	
	{
		//wh[x][set]==0 iff (x in set)
		CC[i][0] = 0;//no cost for going down along the first column[i][0]
		S[i][0] = 3;//direction=3, go down
		
//		k = k+wh[4][A[i]];	//if '-' not in A[i] k++;
//		CC[i][0] = k;
//		if(wh[4][A[i]]) S[i][0] = 2;
//		else S[i][0] = 3;
	}

	for (i=1; i < N+1; i++)	{
		CC[0][i] = i;
		S[0][i] = 1;
	}
	CC[0][0] = 0;
	S[0][0] = 0;

	for (i = 1; i < M+1; i++)	{
		for (j = 1; j < N+1; j++)	{
			temp = CC[i-1][j-1];
			current = CC[i-1][j-1] + wh[B[j]][A[i]];

			if (wh[B[j]][A[i]] == 0) direction = 0;
			else direction = -1;

			if (CC[i][j-1] + g < current)
			{
				current = CC[i][j-1] + g;
				direction = 1;
			}
			
			if (wh[4][A[i]] == 0)
			{
				if (CC[i-1][j] < current)	{
					current = CC[i-1][j];
					direction = 3;
				}
			}
			else 
			{
				if (CC[i-1][j] + g  < current)
				{
               current = CC[i-1][j] + g;
               direction = 2;
            }
			}
			CC[i][j] = current;
			S[i][j] = direction;
		}
	} 

	*pp = p;
	score = CC[M][N];	/*global alignment*/
	
	int minrow = 0;
	int minval = oo;
	for (i = 0; i < M+1; i++) {
		(CC[i][N]<=minval)?(minval=CC[i][N],minrow=i):(0);
	}
	//now we have C[minrow][N] be the minimum score
	score = CC[minrow][N]; //local alignment
	//upadate S[minrow+1][N] to S[M][N];
	for(i=minrow+1; i<M+1; i++) {
		S[i][N]=3;
	}
   printf("score=%d, minrow=%d, totallength=M\n",score,minrow, M);
	for (i = 0; i < M+1; i++) free(CC[i]);
   free(CC);
	return score;
}

int traceback(int **S, int M, int N, int *trace, char * A)
{
	int i, j, k, tt,sc;

	i = M;
	j = N;
	k = 0;
	sc = 0;

	while(!(i<1 && j<1)) {
		tt = S[i][j];
		switch (tt)	{
			case -1:
				trace[k] = -1; k++;
				i = i-1; j = j-1;
				sc++;
				break;
			case 0:
				trace[k] = 0; k++;
				i = i-1; j = j-1;
				break;
			case 1:
				trace[k] = 1; k+=gapB;
				sc++;
				j = j-1;
				break;
			case 3:
				trace[k] = 3; k++;
				i = i-1;
				break;
			case 2:
				trace[k] = 2; k++;
				sc++;
				i = i-1;
				break;
		}
	}
	return (k);
}

int newalign (int *trace, int trace_len, char **aln, int M, char *seq1, int N, char **alnw, int numseq)
{
	int i, j, k, tt, cc, t, b;
	cc = 0;
	t = 0;
	b = 1;

	for (i = trace_len-1; i >=0; i--)
	{
		tt = trace[i];
		switch(tt)
		{
			case -1:
			case 0:
            for (j = 0; j < numseq; j++)
				{
               alnw[j][cc] = aln[j][t];
            }
            alnw[numseq][cc] = seq1[b];
            t++; b++; cc++;
				break;
			case 1: 
				for (j = 0; j < numseq; j++)    
				{
               alnw[j][cc] = 4;
            }
            alnw[numseq][cc] = seq1[b];
            b++; cc++;
            break;
			case 2:
			case 3:
				for (j = 0; j < numseq; j++)
				{
                alnw[j][cc] = aln[j][t];
            }
            alnw[numseq][cc] = 4;
            t++; cc++;
            break;
		}
	}
	return(cc);
}
