/* 
 * fitch.c
 *
 * The fitch algoritm to calculate parsimony 
 *
 * Changes:
 *  04/07/07 - Recursive fitch (jung@ucsd.edu)
 */
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "tree.h"
#include "global.h"
#include "extvab.h"
#include "extfunc.h"


/*
 * Bottom up visit of the nodes and completion of the First and Second set.
 * This is a recursive function in a post order visit.
 */
int forward(TNODE* this_node, char **aln, int seq_len) {
	int i, j;
	int seq_id;
	int counts[5];
	int max_count;

	if(this_node->seqid >= 0)
	{// leaf node, this the base case no need to recurse
	   for (i=0; i<seq_len; i++) 
		{
		   seq_id = this_node->seqid;
			setone(aln[seq_id][i], this_node->first[i]);
		   this_node->cost[i] = 0;
	   }
	}
	else 
	{
	   forward(this_node->lchild, aln, seq_len);
	   forward(this_node->rchild, aln, seq_len);

		// this is an internal node
	   for (i=0; i<seq_len; i++) {
	   max_count = -1;
	   for (j=0; j<5; j++) {
		 // cycle to all the letters, including the gap of the profile
			counts[j] = in(j, this_node->lchild->first[i]) + 
							in(j, this_node->rchild->first[i]);
	      if (counts[j] > max_count)  max_count = counts[j];
	   }

	   for (j=0; j<5; j++) {
	      if (counts[j] == max_count)    setone(j, this_node->first[i]);
	      if (counts[j] == max_count-1)  setone(j, this_node->second[i]);
	   }

	   this_node->cost[i] = this_node->rchild->cost[i] +
	                        this_node->lchild->cost[i] + 2 - max_count;
	   }
	}

	return 0;
}


/*
 * Top down visit of the nodes and completion of the intermidiate sequences
 * of the internal nodes using fitch algorithm.
 * This a recursive function and it is an preorder visit of the tree.
 * We assume first and second are prepopulated.
 */
int backward(TNODE* this_node, int seq_len) {
	int i, j;
	char profile_direct;// the parent profile and first agree
	char profile_eval;// when we don't have a consensus, we need to evaluate
	                  // first and second profiles
	int eval_flag = 0;  // checks whether we are sure of the choices we take

	if (this_node->father == 0)
	{
		memcpy(this_node->profile, this_node->first, seq_len);
	}
	else 
	{
	   for (i=0; i<seq_len; i++) {
	   eval_flag = 0;
		profile_eval = this_node->first[i];
		profile_direct = 0;

	   for (j=0; j<5; j++) 
		{
			if (in(j, this_node->father->profile[i]))
			{
				// the parent has this letter
				if (in(j, this_node->first[i]))
				{
				// the first set agree with the parent profile
					setk(j, profile_direct, this_node->first[i]);
	         }
				else if (in(j, this_node->second[i]))
				{
					setone(j, profile_eval);
		         // not in first but in second
					eval_flag = 1;
	         }
	         else 
				{ // not in first or second set
		         eval_flag = 1;
	         }
	      }
	   }

	   if (eval_flag)	this_node->profile[i] = profile_eval;// first set and parent do not agree
	   else {// first set and parent profile agree
			this_node->profile[i] = profile_direct;
		}
	 	/*
		printf ("The profile of parent is \n");print_profile(this_node->father->profile,seq_len);
	 	printf ("The first of current node is \n");if(!this_node->lchild){printf("LEAF\n");}print_profile(this_node->first, seq_len);
	 	printf ("The second of current node is \n");if(!this_node->lchild){printf("LEAF\n");}print_profile(this_node->second, seq_len);
	 	printf ("The profile of current node is \n");if(!this_node->lchild){printf("LEAF\n");}print_profile(this_node->profile,seq_len);
		*/}
	}

	if (this_node->lchild)     backward(this_node->lchild, seq_len);
	if (this_node->rchild)     backward(this_node->rchild, seq_len);

	return 0;
}

int fixbranchlen(TNODE * this_node) {
	if (this_node->father != 0) {
		this_node->weight = 1.0;
	}
	if (this_node->lchild) fixbranchlen(this_node->lchild);
	if (this_node->rchild) fixbranchlen(this_node->rchild);
	
}

void print_profile(char* profile, int len){
   int i=0;
   int j=0;
 
 	for(j=0; j<5;j++) {
      printf("%2c", na_name[j]);
      for(i=0;i<len;i++) {
         printf("%2d", (1-wh[j][profile[i]]));
      }
      printf("\n");
   }
   printf("\n");
}


