/* 
 * main.c
 *
 * Main for treelign. Given an alignment file, a tree file and a new sequence
 * file, insert the new sequence file into the tree
 *
 * Changes:
 *  04/05/07 - This header created and added comments (jung@ucsd.edu)
 */
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "nuc.h"
#include "tree.h"
#include "weight.h"

char int2char (int tt);
double precision  = 1e-6;

void print_usage() {
	printf("usage: ./FitchAln -i input_aln -f alnformat -t input_tree -o output\n");
	printf("-i input_aln : an incoming multiple sequence alignment (MSA) file\n");
	printf("-t input_tree: an incoming (binary newick) phylogenetic tree, each exterior node of which represents a distinct sequence in the incoming MSA\n");
	printf("-o output:     an outcoming result\n");
	printf("-f alnformat : format of input_aln,  0-clustalw, 1-fasta\n");
	return;
}

/*
 * Process the arguments from main
 */
void initenv(int argc, char **argv, g_options* g_args) {
  int copt;
  int inpseq = 0, tseq = 0, outseq = 0, fseq=0;
  int format = 0, outformat = 0;
  while ((copt=getopt(argc,argv,"i:o:t:a:n:f:F:")) != EOF) {
    switch(copt) {
      case 'i':
        inpseq = 1;
        g_args->align_file = (char*)malloc(sizeof(char)*(strlen(optarg)+1));
        strncpy(g_args->align_file, optarg, strlen(optarg)+1);
        break;
      case 'o':
        outseq =1;
        g_args->out_file = (char*)malloc(sizeof(char)*(strlen(optarg)+1));
        strncpy(g_args->out_file, optarg, strlen(optarg)+1);
        break;
      case 't':
        tseq = 1;
        g_args->tree_file = (char*)malloc(sizeof(char)*(strlen(optarg)+1));
        strncpy(g_args->tree_file, optarg, strlen(optarg)+1);
        break;
      case 'f':
        fseq=1;
        sscanf(optarg, "%d", &format );
        g_args->format = format;
        printf("format: %d\n", format);
        break;
      case '?':  
      default:
        print_usage();
        exit(-1);
    }
  }

  if(inpseq == 0 || outseq == 0 || tseq == 0 ||fseq==0)  {
    print_usage();
    exit(-1);
  }
  return;
}
      

/*
 * Main function for treelign.
 */
int main(int argc, char **argv)
{
	char **aln, **aln2;
	char **name;

	char **tname;
	int *tnode;
	double *tweight;

	TLIST *mtree;
	int *tleaf;

	int length1;
	int temp;
	int score;


		  
	int parent;
	int profile;

	int i,j,k;
	int numseq;
	int length;   
	int numnode, numtnode;
	int check;
	long cost;
	FILE *fp;

	TNODE* root;
	TNODE* best_node;
	TNODE* new_node;      // store the sequence to inserted in here
	int best_score;

	g_options g_args = {0};
	char* tree_string = 0;

	int M,N;

	initenv(argc, argv, &g_args);

	//read a clustalw format
	printf("Reading Multiple Sequence Alignment ...\n");
	fp = fopen(g_args.align_file, "r");
	if (!fp) {
	 fprintf(stderr, "ERROR: Unable to open file %s\n", g_args.align_file);
	 exit(-1);
	}
	//space for the alignment
	aln  = (char**)malloc(MAXSEQ * sizeof(char*));
	name = (char**)malloc(MAXSEQ * sizeof(char*));
	for (i = 0; i < MAXSEQ ; i++)   {
	 aln[i] = (char*)malloc(MAXNUM * sizeof (char));
	 name[i] = (char*)malloc(MAXNAME * sizeof (char));
	}
	string iformat = "";
	if(g_args.format == 0) {
		iformat = "clustalw";
		check = readalign(aln, name, &length, &numseq, fp);
	}else if(g_args.format == 1) {
		check = readfasta(aln, name, &length, &numseq, fp);
		iformat = "fasta";
	}	

	printf ("Done, the number of species is %d\n", numseq);
	fclose (fp);

	//read a Newick tree;
	printf("Reading Newick Tree ...\n");
	fp =fopen(g_args.tree_file, "r");
	if (!fp) {
	 fprintf(stderr, "ERROR: Unable to open file %s\n", g_args.tree_file);
	 exit(-1);
	}
  	//space for the tree;
	tname = (char**)malloc((numseq+1)*sizeof(char*));
	tnode = (int*)malloc((5*numseq)*sizeof(int));
	tweight = (double*)calloc(4*numseq, sizeof(double));
	for (i = 0; i < numseq+1; i++)    {
	 tname[i] = (char *) malloc (MAXNAME * sizeof (char));
	}
	numnode = readtree(tname, tnode, tweight, numseq, fp);
	fclose (fp);
	printf ("Done, the number of nodes is %d\n", numnode);
	

	
		

	//vertical compress the input aln[][], removing cloumns that are only composed of dashes
	char** aln_tmp;
	int length_tmp=0;
	bool* dashcolumn;
	aln_tmp = (char**)malloc(numseq*sizeof(char*));
	for(i=0; i<numseq; i++){
		aln_tmp[i] = (char*)malloc(length*sizeof(char));
	}
	dashcolumn = (bool*)malloc(length*sizeof(bool));
	for(j=0;j<length;j++)	dashcolumn[j]=true;

	for(j=0;j<length;j++)
		for(i=0; i<numseq; i++)
		if(aln[i][j]!=Char2int('-'))
			dashcolumn[j]=false;
	for(j=0;j<length;++j){
		if(!dashcolumn[j]) {//the j-th column of aln[][] can not be compressed
			for(i=0;i<numseq;++i){
				aln_tmp[i][length_tmp] = aln[i][j];
			}
			length_tmp++;
		}
	}

	aln2 = (char**)malloc(numseq * sizeof(char*));
	for (i = 0; i < numseq; i++) {
	 aln2[i] = (char*)malloc(length_tmp * sizeof(char));
	}
	for (i = 0; i < numseq; i++){ 
		for (k = 0; k < numseq; k++){
			if (strcmp(tname[i], name[k])==0)
				for (j=0; j< length_tmp; j++)
					 aln2[i][j] = aln_tmp[k][j];
		}
	}
	if(length != length_tmp) printf("The MSA is compressed by removing columns that are totally composed of dashes (length %d-->%d).\n", length, length_tmp);
	length = length_tmp;
	
	//free memory
	for(i=0;i<numseq; ++i){
		free((void*)aln_tmp[i]);
	}
	free((void**)aln_tmp);
	free((void*)dashcolumn);


	//need check name[] order and tname[]
	//ADDNUM = 100
	mtree = (TLIST*)malloc((2*numseq + ADDNUM) * sizeof(TLIST));
	tleaf = (int*)malloc(numseq * sizeof(int));

	//buildtree;
	numtnode = buildtree(&root, mtree, tleaf, tnode, numseq, numnode, tweight);
	root->father = 0;          // mark this as the root
	//assign nodeindx in broadth first search order
	assignnodeindxBFS(root,0, tname);

	int index = 0;
	inorder_allocate(root, length);
	forward(root, aln2, length);
	backward(root, length);
	cost = 0;
	for (i=0; i<length; i++) {
	   cost += root->cost[i];
	}
	printf("Fitch score of the input tree is %ld\n", cost);
	
	
	//output the new aln and new tree.
	fp =fopen(g_args.out_file, "w");
	if(fp == NULL)	{
		printf("Can not open outputfile %s\n", g_args.out_file);
		exit(-1);
	}
	fprintf(fp, "#############FIRST PART#####################################\n");
	printsubtreeBFS(fp, tname, root);
	fprintf(fp, "\n#############SECOND PART####################################\n");
	
	fprintf(fp, "F.S.M  \t");
	for(i=0; i<length; i++) {
		fprintf(fp, "%d\t", i);
	}
	fprintf(fp, "sum\n");
	printfitchBFS(fp, length, root);
	
	fprintf(fp, "\n");
	fclose(fp);

	//free space;
	for (i= 0; i < MAXSEQ; i++) {
	   free((void*)aln[i]);
	   free((void*)name[i]);
	}

	free((void**)aln);
	free((void**)name); 
	free((void*)tnode);
	free((void*)tleaf);


	for (i=0; i < numseq; i++) {
	   free((void*)tname[i]);
	   free((void*)aln2[i]);
	}

	free((void*)mtree); 
	free((void**)tname);
	free((void**)aln2);

	// free global_args file
	free(g_args.align_file);
	free(g_args.tree_file);
	free(g_args.out_file);
	return 0;        
}

