#include <stdio.h>
#include <string>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include "tree.h"

int readalign (char **aln, char **name, int *length, int *numseq, FILE *fp);
int readtree (char **tname, int *tnode, double *tweight, int numseq, FILE *fp);
int readfasta (char **aln, char **name, int *length, int *numseq, FILE *fp);
int readoneseq (char *seq, char *name, FILE *fp);
char Char2int(char c);


/*
 * Reads in the tree from the file. We assume that the tree is in Newick 
 * format. 
 */
int readtree(char** tname, int* tnode, double* tweight, int numseq, FILE* fp) {
	char str[500], temp[200], temp2[200];
	char pre='x';	//save the previous character that was read.
	double score;
	int n = 0;
	int m = 0;
	int p = 0; 
	int i, j, k;
	int t,s;
	int finished;
	   
	finished = 0;
	while (fgets(str, 450, fp) && !finished) 
	{
		if (str[0] != ' ') 
		{
			for (j=0; j<strlen(str); j++)   
			{
//				printf ("%c", str[j]);
			   switch (str[j]) {
			      case '(':
			      k = 0;
			      tnode[n] = -1;
			      n++;pre='(';
			      break;
			      case ')':
			      if (k>0) {
			         temp[k] = '\0';
						score = atof(temp);
			         tweight[p] = score; 
//			         printf("Score = %lf\n", score) ;
						k = 0;
			         p++;
			      }
			      k=0;
			      tnode[n] = -2;	//tnode[i] = -1 when we enter the subtree, tnode[i]=-2 when we finish traversing the substree
			      n++;	pre=')';
			      break;
			      case ':':       
					{
						if(pre!=')'){
					      temp[k] = '\0';
					      strcpy(tname[m], temp);
					      k = 0;
					      tnode[n] = m;
					      n++;
					      m++;
					   }
					   k=0;	pre=':';
					   break;  
					}
			      case '\n':
			      case '\t':
					pre='\n';
			      break;
			      case ';':
			      finished = 1;
					pre=';';
			      break;
			      case ',':
			      if (k>0) {
			         temp[k] = '\0';
			         score = atof(temp);
			         tweight[p] = score;
//			         printf("Score = %lf \n", score);
						k = 0;
			         p++;
			      }
			      k = 0;pre=',';
			      break;
			      default:
			      temp[k]=str[j];
			      k++;
			      break;
			      }
			  }
		}//end if(str[0] != ' ')
	}//end while(fgets(str, 450, fp) && !finished)
	if (m != numseq) {
	printf("Sequence number is not consistent with the alingment m = %d: numseq = %d\n", m, numseq);
	}
//	printf ("Exit reading tree, n=%d\n", n);
	return (n);

}

//read alignment in clustal format                
int readalign (char **aln, char **name, int *length, int *numseq, FILE *fp) 
{
	char str[500];
	int n=0;
	int m=0;
	char main_str[500];
	int i,j, k;
	int check = -1;
	i = 0;
	while (fgets(str, 490, fp))  
	{
		if (strncmp(str,"CLUS",4)==0) {} 
		else if(isalnum(str[0])) 
		{
			//read one block, which contains a substring of all species.
         i = 0;

			//j marks the starting position for the substrings of the current 
			//block.
         j = n;
         while (isalnum(str[0]))
			{
            n = j;
            sscanf(str, "%s %s", name[i], main_str);
            for (k=0; k<strlen(main_str); k++) 
				{
               if (islower(main_str[k]))   {
                     aln[i][n++] = Char2int(main_str[k]);
               } else if (isupper(main_str[k]))  {
                     aln[i][n++] = Char2int(tolower(main_str[k]));
               } else if (main_str[k] == '-')  {
                     aln[i][n++] = 4;
               }
            }
            i++;
            if (fgets(str, 490, fp))        {}
            else {str[0] = ' ';};
         }
         *numseq = i;
      }
	}

	*length = n;
	check++;
	return (check);
}

//read alignment in fasta format
int readfasta (char **aln, char **name, int *length, int *numseq, FILE *fp)     
{
	char str[MAXNUM+50];
	char main_str[MAXNUM];
	int n = 0;
	int m = 0;
	int i, j, k;
	i = 0;
	while (fgets(str, MAXNUM, fp))  {
		if (strcmp(str, "\n") == 0) {
			continue;
		}
		if (str[0] == '>')      {
			sscanf (str, ">%s", name[i]);
		} else {
			sscanf (str, "%s", main_str);
			n = 0;
			for (k=0; k< strlen(main_str); k++) {
				if (main_str[k] >= 'a' && main_str[k] <= 'z')   {
					aln[i][n++] = Char2int(main_str[k]);
				} else if (main_str[k] >= 'A' && main_str[k] <= 'Z')  {
					aln[i][n++] = Char2int(main_str[k] - 'A' + 'a');
				} else if (main_str[k] == '-')  {
					aln[i][n++] = 4;
				}
			}
			i++;
		}
	}
	*numseq = i;
	*length = n;
	return (0);
}

int readoneseq (char *seq, char *name, FILE *fp){
	char str[500];
	char main_str[450];
	int n = 0;
	int k;
	while(fgets(str, 450, fp)){
		if(str[0] == '>') sscanf(str, ">%s", name);
		else {
			string s(str);
			for(k=0; k<s.size(); ++k)
				if(isalpha(s[k])) seq[++n] = Char2int(tolower(s[k]));
		}
	}
	return (n);
}


char Char2int(char c)  
{
	switch(c){
	case 'g' :	return 0;
	case 'a' :	return 1;
	case 'c' :	return 2;
	case 't' :	return 3;
	case 'u' :  return 3;
	case '-' :	return 4;
	default  :	break;
	}
	//	printf ("Can not find letter: %c in alphabeta {gact-}\n", c);
	return(rand()%4);
}


