/* 
 * tree.h
 *
 * Operations and structures related to the tree 
 *
 * Changes:
 *  04/05/07 - Header creation (jung@ucsd.edu)
 */
#ifndef _TREE_H_
#define _TREE_H_
#include <stdio.h>
#include <string.h>
#include <string>
#include <math.h>
#include <stdlib.h>
#include "extvab.h"
#include "extfunc.h"

#include "global.h"
#include <queue>
#include <vector>
#include <assert.h>
#define sz(x) (int)(x).size()
#define FOR(i,a,b) for(int i=a; i<b; ++i)
#define PB push_back
#define LIDEBUG
using namespace std;

/*
 * Structure definitions
 */
class TNODE{
	TNODE()	{profile = first = second = NULL; cost = NULL;}
public:
	char* profile;
	char* first;
	char* second;
  int* cost;

  int totalcost;
  int seqid;	//seqid>=0 iff it's a leaf node
  double weight;

  struct TNODE* lchild;
  struct TNODE* rchild;
  struct TNODE* father;

  int level;
  int numleaf;
  int nodeindx;
  /*align sequence with profile of this branch using Profile Sequence Alignment*/
  int psa;
  int totalnumleaf; /* the total number of descendent leaf nodes*/
  double dist2givennode;  /* the distance from current node to a given node*/
  int score; /*MP score of aligning the input sequece and the branch profile*/
};

class LNODE {
	LNODE()	{prev=NULL; data=NULL;}
public:  
	LNODE* prev;
  TNODE* data;
};


class TLIST{
public:
	int totalcost;
	int seqid;
	double weight;
	
	//position of left child, right child, parent, and sibling in mtree
	int left;
	int right;
	int father;
	int brother;

	int flag;
	int level;
	int numleaf;
};


/*
 * Function prototypes
 */
void printsubtreeBFS(FILE* fout, char** names, TNODE* current);
void printfitchBFS(FILE* fout, int length, TNODE* current);
void assignnodeindxBFS(TNODE* current, int indx, char** tname);


int readtree(char** tname, int* tnode, double* tweight, int numseq, FILE *fp);
int buildtree(TNODE** root, TLIST* mtree, int* tleaf, int* tnode, int numseq, int numnode, double * tweight);
int removeleaf (TLIST* mtree, int numseq, int numnode, int n);
void insert_node(TNODE* target, TNODE* insert, int length);
void print_newlick(FILE* fout, char** names, TNODE* current);

int forward(TNODE* root, char** aln, int length);
int backward(TNODE* root, int length);
int fixbranchlen(TNODE * root);


void inorder_allocate(TNODE* current, int length, int allocate_profile=1);
void inorder_free(TNODE* current);
int postorder_align(TNODE* current, char* profile, int prof_len, 
                    char* sequence, int seq_len, int** grid, TNODE** best,
                    int best_score, int** CC);

int continue_align (TNODE* current, char* profile, int prof_len,
                    char* sequence, int seq_len, int ** grid,
                    int best_score, int** CC);

TNODE* searchleafbyseqid(TNODE* current, const int seqid);
TNODE* searchleafbyname(TNODE* current, const char* queryname, char** names);
//int getNode(TNODE* current, string target_name,char** tname);
/*get all the leaf node and store them in leafvec*/
void getLeafVec(TNODE* curNode, vector<TNODE*> &leafvec);
void print_profile(char* profile, int len);
double caldist(char * s1, char * s2, int length);
void assignweight(TNODE * root, double * tweight, int &index);

void caltotalnumleaf(TNODE* current);
void decide(TNODE* current, int num);
void setpsa(TNODE* current);
/*for each node, compute its distance to a given node (save in dist2givennode)
  and set psa if the distance is less than searchradius, return the number of psa=1 */
int cmpdist_setpsa(TNODE* current, int direction);
#endif
